@@ -1,14 +1,14 @@
 ;;; reaper.el --- Interact with Harvest time tracking app  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2019  Thomas Fini Hansen
+;; Copyright (C) 2019-2026  Thomas Fini Hansen
 
 ;; Author: Thomas Fini Hansen <xen@xen.dk>
 ;; Created: August 11, 2019
-;; Package-Version: 1.5.0
-;; Package-Revision: 7515ef42e0bf
+;; Package-Version: 1.6.0
+;; Package-Revision: v1.6.0-0-g71c0fcb54a13
 ;; Package-Requires: ((emacs "26.2"))
 ;; Keywords: tools
-;; Url: https://github.com/xendk/reaper
+;; URL: https://github.com/xendk/reaper
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -27,12 +27,13 @@
 
 ;; Interactive tool for tracking time with Harvest.
 
+
+;;; Code:
 (require 'cl-lib)
 (require 'json)
 (require 'url)
 
 ;; Declare what we're using from elsewhere.
-(declare-function ivy-read "ivy")
 (defvar calc-eval-error)
 
 (defgroup reaper nil
@@ -61,7 +62,6 @@
 (defconst reaper--time-regexp
   (rx (optional (group-n 1 (optional digit)) ":") (group-n 2 (one-or-more digit))))
 
-;;; Code:
 (defvar reaper-buffer-name " *Reaper*"
   "Name for Reaper buffer.")
 
@@ -78,6 +78,27 @@ supplied note. If one of these functions returns a cons
 of (project_id . task_id), those are used and the rest are not
 called.")
 
+(defvar reaper-description-history nil
+  "History of descriptions.")
+
+(defvar reaper-date-history nil
+  "History of dates.")
+
+(defvar reaper-time-history nil
+  "History of time inputs.")
+
+(defvar reaper-project-history nil
+  "History of project selections.")
+
+(defvar reaper-task-history nil
+  "History of task selections.")
+
+(put 'reaper-description-history 'history-length 100)
+(put 'reaper-date-history 'history-length 10)
+(put 'reaper-time-history 'history-length 10)
+(put 'reaper-project-history 'history-length 10)
+(put 'reaper-task-history 'history-length 10)
+
 (defvar-local reaper-user-id nil
   "Cached id of user.")
 
@@ -91,8 +112,9 @@ called.")
   "Cache of projects and tasks.")
 
 (defvar-local reaper-last-marked-entry nil
-  "ID of the time entry point was on before last fetch. Used to
-relocate point to the same entry after redisplaying.")
+  "ID of the time entry point was on before last fetch.
+
+Used to relocate point to the same entry after redisplaying.")
 
 (defvar-local reaper-running-timer nil
   "Currently running timer.")
@@ -197,8 +219,8 @@ relocate point to the same entry after redisplaying.")
 
 (define-derived-mode reaper-mode tabulated-list-mode "Reaper"
   "Major mode for Reaper buffer.
-\\<reaper-mode-map>
-"
+
+\\<reaper-mode-map>"
   :group 'reaper
   :syntax-table nil
   :abbrev-table nil
@@ -232,7 +254,7 @@ relocate point to the same entry after redisplaying.")
   (interactive)
   (reaper--check-credentials)
   (reaper-with-buffer
-   (pop-to-buffer (current-buffer))))
+   (pop-to-buffer (current-buffer) '(display-buffer-reuse-window (reusable-frames . visible)) t)))
 
 (defun reaper-get-running-timer-note ()
   "Return the note (description) of the current running timer.
@@ -245,7 +267,9 @@ If no timer is running, return nil."
          (reaper-entry-notes entry))))))
 
 (defun reaper-insert-project-id ()
-  "Prompt for a project and insert the id of the selected project into the current buffer."
+  "Insert project id into the current buffer.
+
+Prompt for a project and insert the id of the selected project."
   (interactive)
   (reaper-ensure-project-tasks)
   (let ((project (reaper-with-buffer (reaper-read-project (reaper-project-id (reaper-get-head-project))))))
@@ -253,7 +277,9 @@ If no timer is running, return nil."
       (insert (number-to-string (reaper-project-id project))))))
 
 (defun reaper-insert-task-id ()
-  "Prompt for a project and task, and insert the id of the selected task of project into the current buffer."
+  "Insert task id into the current buffer.
+
+Prompt for a project and task, and insert the id of the selected task."
   (interactive)
   (reaper-ensure-project-tasks)
   (let* ((project (reaper-with-buffer (reaper-read-project (reaper-project-id (reaper-get-head-project)))))
@@ -319,16 +345,19 @@ If no timer is running, return nil."
 
 (defun reaper-clear-project-tasks ()
   "Clear cached projects and tasks."
+  (declare (modes reaper-mode))
   (interactive)
   (setq reaper-project-tasks nil))
 
 (defun reaper-ensure-project-tasks ()
   "Ensure that we have project and tasks fetched."
-  (unless (bound-and-true-p reaper-project-tasks)
-    (reaper-refresh-project-tasks)))
+  (reaper-with-buffer
+   (unless (bound-and-true-p reaper-project-tasks)
+     (reaper-refresh-project-tasks))))
 
 (defun reaper-refresh-project-tasks ()
   "Fetch projects and tasks from Harvest."
+  (declare (modes reaper-mode))
   (interactive)
   (reaper-with-buffer
    (unless reaper-project-tasks
@@ -356,6 +385,7 @@ If no timer is running, return nil."
 
 (defun reaper-refresh-buffer ()
   "Refresh Reaper buffer."
+  (declare (modes reaper-mode))
   (interactive)
   (unless (and (bound-and-true-p reaper-timeentries) (not reaper-timeentries-loading))
     (reaper-refresh-entries))
@@ -363,7 +393,6 @@ If no timer is running, return nil."
 
 (defun reaper-update-buffer ()
   "Update Reaper buffer."
-  (interactive)
   (setq mode-name (concat "Reaper " reaper-date))
   (tabulated-list-init-header)
   (tabulated-list-print t)
@@ -401,7 +430,7 @@ which case the month is the previous.
 Alternatively +<days>/-<days> can be used to move X days
 forward/back from reaper-date."
   (interactive)
-  (reaper--goto-date (read-string "Goto date: ")))
+  (reaper--goto-date (read-string "Goto date: " nil 'reaper-date-history)))
 
 (defun reaper-goto-date+1 ()
   "Go a day forward."
@@ -416,6 +445,7 @@ forward/back from reaper-date."
 (defun reaper-start-timer ()
   "Start the timer at point.
 Stops any previously running timers."
+  (declare (modes reaper-mode))
   (interactive)
   (when (tabulated-list-get-id)
     (reaper-api "PATCH" (format "time_entries/%s/restart" (tabulated-list-get-id)) nil "Started timer")
@@ -438,7 +468,7 @@ Stops any previously running timers."
   "Create a new running timer."
   (interactive)
   (reaper-ensure-project-tasks)
-  (let* ((notes (read-string "Description: "))
+  (let* ((notes (read-string "Description: " nil 'reaper-description-history))
          (autofile (run-hook-with-args-until-success 'reaper-autofile-functions notes))
          (project
           (or (and (consp autofile) (reaper-get-project (car autofile)))
@@ -457,12 +487,13 @@ Stops any previously running timers."
 
 (defun reaper-delete-entry ()
   "Delete time entry at point."
+  (declare (modes reaper-mode))
   (interactive)
   (reaper-with-selected-entry
-   (when (yes-or-no-p (format "Are you sure you want to delete \"%s - %s: %s"
+   (when (yes-or-no-p (format "Are you sure you want to delete \"%s\" in \"%s - %s\"?"
+                              (reaper-entry-notes entry)
                               (reaper-entry-project entry)
-                              (reaper-entry-task entry)
-                              (reaper-entry-notes entry)))
+                              (reaper-entry-task entry)))
      ;; Go forward a line, so tabulated-list-mode has an entry to
      ;; stick to.
      (forward-line 1)
@@ -480,12 +511,13 @@ Stops any previously running timers."
 
 (defun reaper-edit-entry ()
   "Edit entry at point."
+  (declare (modes reaper-mode))
   (interactive)
   (reaper-ensure-project-tasks)
   (reaper-with-selected-entry
    (let* ((project (reaper-read-project (reaper-entry-project-id entry)))
           (task-id (reaper-read-task project (reaper-entry-task-id entry)))
-          (notes (read-string "Description: " (reaper-entry-notes entry)))
+          (notes (read-string "Description: " (reaper-entry-notes entry) 'reaper-description-history))
           (harvest-payload (make-hash-table :test 'equal)))
      (puthash "project_id" (cdr (assoc :id project)) harvest-payload)
      (puthash "task_id" task-id harvest-payload)
@@ -495,6 +527,7 @@ Stops any previously running timers."
 
 (defun reaper-edit-entry-project ()
   "Edit project of entry at point."
+  (declare (modes reaper-mode))
   (interactive)
   (reaper-ensure-project-tasks)
   (reaper-with-selected-entry
@@ -515,6 +548,7 @@ Stops any previously running timers."
 
 (defun reaper-edit-entry-task ()
   "Edit task of entry at point."
+  (declare (modes reaper-mode))
   (interactive)
   (reaper-ensure-project-tasks)
   (reaper-with-selected-entry
@@ -527,10 +561,11 @@ Stops any previously running timers."
 
 (defun reaper-edit-entry-description ()
   "Edit description of entry at point."
+  (declare (modes reaper-mode))
   (interactive)
   (reaper-ensure-project-tasks)
   (reaper-with-selected-entry
-   (let* ((notes (read-string "Description: " (reaper-entry-notes entry)))
+   (let* ((notes (read-string "Description: " (reaper-entry-notes entry) 'reaper-description-history))
           (harvest-payload (make-hash-table :test 'equal)))
      (puthash "notes" notes harvest-payload)
      (reaper-api "PATCH" (format "time_entries/%s" (reaper-entry-id entry)) harvest-payload "Updated entry")
@@ -538,11 +573,12 @@ Stops any previously running timers."
 
 (defun reaper-edit-entry-time ()
   "Edit time of entry at point."
+  (declare (modes reaper-mode))
   (interactive)
   (reaper-with-selected-entry
    ;; If the timer is running add the time since the data was fetched.
    (let* ((time (reaper--hours-to-time (reaper--hours-accounting-for-running-timer entry)))
-          (new-time (reaper--time-to-hours-calculation (read-string "New time: " time)))
+          (new-time (reaper--time-to-hours-calculation (read-string "New time: " time 'reaper-time-history)))
           (harvest-payload (make-hash-table :test 'equal)))
      (puthash "hours" new-time harvest-payload)
      (reaper-api "PATCH" (format "time_entries/%s" (reaper-entry-id entry)) harvest-payload "Updated entry")
@@ -558,7 +594,7 @@ Stops any previously running timers."
                            reaper-project-tasks))
          (default (reaper-get-project default))
          (default-option (when default (concat "[" (reaper-project-code default) "] " (reaper-project-name default))))
-         (project (cdr (assoc (reaper--completing-read "Project: " projects default-option) projects))))
+         (project (cdr (assoc (reaper--completing-read "Project: " projects default-option 'reaper-project-history) projects))))
     project))
 
 (defun reaper-read-task (project &optional default)
@@ -567,83 +603,77 @@ Returns task id."
   (let*
       ((tasks (mapcar (lambda (task) (cons (cdr task) (car task))) (reaper-project-tasks project)))
        (default (when default (cdr (assoc default (reaper-project-tasks project)))))
-       (task-id (cdr (assoc (reaper--completing-read "Task: " tasks default) tasks))))
+       (task-id (cdr (assoc (reaper--completing-read "Task: " tasks default 'reaper-task-history) tasks))))
     task-id))
 
+(defmacro reaper--with-envelope (method path payload &rest body)
+  "Run BODY with HTTP envelope.
+
+Prepare call to Harvest PATH with HTTP METHOD and PAYLOAD encoded as
+JSON."
+  `(let* ((url-request-method ,method)
+          (url-mime-language-string nil)
+          (url-mime-encoding-string nil)
+          (url-mime-accept-string "application/json")
+          (url-user-agent "Xen's Emacs client (fini@reload.dk)")
+          (url-request-data (if (or (string-equal method "POST")
+                                    (string-equal method "PATCH"))
+                                (encode-coding-string (json-encode ,payload) 'utf-8)
+                              nil))
+          (request-url (format "https://api.harvestapp.com/v2/%s" ,path))
+          (url-request-extra-headers
+           `(("Content-Type" . "application/json")
+             ("Authorization" . ,(concat "Bearer " reaper-api-key))
+             ("Harvest-Account-Id" . ,reaper-account-id))))
+     ,@body))
+
 (defun reaper-api (method path payload completion-message)
   "Make an METHOD call to PATH with PAYLOAD and COMPLETION-MESSAGE."
   (reaper--check-credentials)
-  (let* ((url-request-method method)
-         ;;(url-set-mime-charset-string)
-         (url-mime-language-string nil)
-         (url-mime-encoding-string nil)
-         (url-mime-accept-string "application/json")
-         (url-personal-mail-address nil)
-         (url-request-data (if (or (string-equal method "POST")
-                                   (string-equal method "PATCH"))
-                               (encode-coding-string (json-encode payload) 'utf-8)
-                             nil))
-         (request-url (format "https://api.harvestapp.com/v2/%s" path))
-         (url-request-extra-headers
-          `(("Content-Type" . "application/json")
-            ("Authorization" . ,(concat "Bearer " reaper-api-key))
-            ("Harvest-Account-Id" . ,reaper-account-id)
-            ("User-Agent" . "Xen's Emacs client (fini@reload.dk)"))))
-    (with-temp-buffer
-      (reaper-url-insert-file-contents request-url)
-      (goto-char (point-min))
-      (message "%s" completion-message)
-      ;; Ensure JSON false values is nil.
-      (defvar json-false)
-      (let ((json-false nil))
-        (json-read)))))
+  (reaper--with-envelope
+   method path payload
+   (with-temp-buffer
+     (let ((data (reaper-url-insert-file-contents request-url)))
+       (message "%s" completion-message)
+       data))))
 
-(defun reaper-url-insert-file-contents (url &optional visit beg end replace)
+(defun reaper--handle-response (buffer url)
+  "Handle response.
+
+BUFFER is the buffer to insert response into, URL is the URL fetched.
+
+Returns the decoded JSON data."
+  (when (fboundp 'url-http--insert-file-helper)
+    (url-http--insert-file-helper buffer url))
+  (url-insert-buffer-contents buffer url)
+  (goto-char (point-min))
+  ;; Ensure JSON false values is nil.
+  (defvar json-false)
+  (let ((json-false nil))
+    (json-read)))
+
+(defun reaper-url-insert-file-contents (url)
   "Quiet version of `url-insert-file-contents'.
 URL, VISIT, BEG, END and REPLACE is the same as for
 `url-insert-file-contents'."
+  ;; The extra t to `url-retrieve-synchronously' is the quiet bit.
   (let ((buffer (url-retrieve-synchronously url t)))
     (unless buffer (signal 'file-error (list url "No Data")))
-    (when (fboundp 'url-http--insert-file-helper)
-      ;; XXX: This is HTTP/S specific and should be moved to url-http
-      ;; instead.  See bug#17549.
-      (url-http--insert-file-helper buffer url visit))
-    (url-insert-buffer-contents buffer url visit beg end replace)))
+    (reaper--handle-response buffer url)))
 
 (defun reaper-api-async (method path payload callback)
-  "Make an asynchronous METHOD call to PATH with PAYLOAD and call CALLBACK on completion."
+  "Make an asynchronous HTTP call.
+
+Make a METHOD call to PATH with PAYLOAD and call CALLBACK on completion."
   (reaper--check-credentials)
-  (let* ((url-request-method method)
-         ;;(url-set-mime-charset-string)
-         (url-mime-language-string nil)
-         (url-mime-encoding-string nil)
-         (url-mime-accept-string "application/json")
-         (url-personal-mail-address nil)
-         (url-request-data (if (or (string-equal method "POST")
-                                   (string-equal method "PATCH"))
-                               (encode-coding-string (json-encode payload) 'utf-8)
-                             nil))
-         (request-url (format "https://api.harvestapp.com/v2/%s" path))
-         (url-request-extra-headers
-          `(("Content-Type" . "application/json")
-            ("Authorization" . ,(concat "Bearer " reaper-api-key))
-            ("Harvest-Account-Id" . ,reaper-account-id)
-            ("User-Agent" . "Xen's Emacs client (fini@reload.dk)"))))
-    (url-retrieve request-url
-                  #'(lambda (&rest ignored)
-                      (let ((async-buffer (current-buffer)))
-                        (with-temp-buffer
-                          (when (fboundp 'url-http--insert-file-helper)
-                            ;; XXX: This is HTTP/S specific and should be moved to url-http
-                            ;; instead.  See bug#17549.
-                            (url-http--insert-file-helper async-buffer request-url))
-                          (url-insert-buffer-contents async-buffer request-url)
-                          (goto-char (point-min))
-                          ;; Ensure JSON false values is nil.
-                          (defvar json-false)
-                          (let ((json-false nil))
-                            (funcall callback (json-read))))))
-                  nil t)))
+  (reaper--with-envelope
+   method path payload
+   (url-retrieve request-url
+                 #'(lambda (&rest _ignored)
+                     (let ((async-buffer (current-buffer)))
+                       (with-temp-buffer
+                         (funcall callback (reaper--handle-response async-buffer request-url)))))
+                 nil t)))
 
 (defun reaper-alist-get (symbols alist)
   "Look up the value for the chain of SYMBOLS in ALIST."
@@ -712,7 +742,7 @@ Will create it if it doesn't exist yet."
          (list nil (vector "Total" "" (reaper--hours-to-time reaper-total-hours) ""))))))))
 
 (defun reaper--highlight-running-and-move-point ()
-  "Highlight the currently running timer, and move point to the last selected entry."
+  "Highlight running timer, and move point to the last selected entry."
   (let ((move-to))
     (save-excursion
       (goto-char (point-min))
@@ -751,7 +781,10 @@ A calculation is a number of time strings (as parsed by
 (defun reaper--time-to-hours (time)
   "Convert TIME to hours.
 TIME is in HH:MM or MM format. Returns a float."
-  (when (string-match (rx bos (regexp reaper--time-regexp) eos) time)
+  ;; Using this concat rather than (rx ... (regexp
+  ;; reaper--time-regexp)), as the regexp form is only supported in
+  ;; Emacs 27+.
+  (when (string-match (concat "^" reaper--time-regexp "$") time)
     (let ((hours (match-string-no-properties 1 time))
           (minutes (string-to-number (match-string-no-properties 2 time))))
       (+ (if hours (string-to-number hours) 0) (/ (float minutes) 60)))))
@@ -776,7 +809,7 @@ many days from reaper-date."
                                                       (+ (string-to-number (nth 2 current-reaper-date)) days-offset)
                                                       (string-to-number (nth 1 current-reaper-date))
                                                       (string-to-number (nth 0 current-reaper-date)))))
-      (let* ((parts (reverse (split-string date-string "[-\.]")))
+      (let* ((parts (reverse (split-string date-string "[-.]")))
              (day (string-to-number (nth 0 parts)))
              (month (and (nth 1 parts) (string-to-number (nth 1 parts))))
              (year (and (nth 2 parts) (string-to-number (nth 2 parts))))
@@ -814,17 +847,10 @@ many days from reaper-date."
                                                    reaper-fetch-time)) 60) 60))
         0)))
 
-(defun reaper--completing-read (prompt options default)
+(defun reaper--completing-read (prompt options default &optional history)
   "Complete with PROMPT, with OPTIONS and having DEFAULT.
-Wraps `completing-read', or `ivy-read', in order to sort options
-in last used order when using ivy."
-  (if (eq completing-read-function 'ivy-completing-read)
-      (ivy-read
-       prompt options
-       :require-match t
-       :preselect default
-       :def default)
-    (completing-read prompt options nil t nil nil default)))
+Optional HISTORY argument tracks completion history."
+  (completing-read prompt options nil t nil history default))
 
 (defun reaper--last-used (project task-id)
   "Save PROJECT and TASK-ID as last used."
